;;==========================================================================;;
;; Joe Zbiciak's Song Player                                                ;;
;; Copyright 2000, Joe Zbiciak.                                             ;;
;; http://www.primenet.com/~im14u2c/intv/                                   ;;
;;==========================================================================;;

;* ======================================================================== *;
;*  This program is free software; you can redistribute it and/or modify    *;
;*  it under the terms of the GNU General Public License as published by    *;
;*  the Free Software Foundation; either version 2 of the License, or       *;
;*  (at your option) any later version.                                     *;
;*                                                                          *;
;*  This program is distributed in the hope that it will be useful,         *;
;*  but WITHOUT ANY WARRANTY; without even the implied warranty of          *;
;*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU       *;
;*  General Public License for more details.                                *;
;*                                                                          *;
;*  You should have received a copy of the GNU General Public License       *;
;*  along with this program; if not, write to the Free Software             *;
;*  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.               *;
;* ======================================================================== *;
;*                   Copyright (c) 2000, Joseph Zbiciak                     *;
;* ======================================================================== *;

        ROMW    10
        ORG     $5000

; Magic memory locations
VBLANK  EQU     $20             ; Vertical-blank Handshake
COLSTK  EQU     $21             ; Color-stack/FGBG switch
CS0     EQU     $28             ; Color Stack 0
CS1     EQU     $29             ; Color Stack 1
CS2     EQU     $2A             ; Color Stack 2
CS3     EQU     $2B             ; Color Stack 3
CB      EQU     $2C             ; Color for border
ISRVEC  EQU     $100            ; ISR jump vector
PSV1	EQU	$300 ;..$305	; PSG1 state save area 
PSV2	EQU	$306 ;..$30B	; PSG2 state save area 
LAST1	EQU	$30C		; Last record pointer for stream 1
LAST2	EQU	$30D		; Last record pointer for stream 2
PSG     EQU     $35F            ; PSG pointer
PSG1    EQU     $1F0            ; PSG base address
PSG2    EQU     $0F0            ; PSG base address

SNDSTRM EQU     $340 ;..$347    ; $340..$347:  Sound stream info

ROMHDR:
        BYTE    $01,$00         ; Movable object data (ignored)
        WORD    $0000           ; RTAB (ignored)
        WORD    INIT            ; Program start address (ignored)
        WORD    $0000           ; Background graphics
        WORD    ROMHDR          ; Card table -- stored above in header. :-)
        WORD    TITLE           ; Title string:  'TETRIS'

        DECLE   $3C0            ; run code after title, clicks off, Inty 2 fix
        BYTE    $00             ; -> to STIC $32
        BYTE    $00             ; 0 = color stack, 1 = f/b mode
        BYTE    0, 0, 0, 0      ; color stack elements 1 - 4
        BYTE    $00             ; border color

TITLE:  BYTE    100, "Song Player", 0
	; Intercept EXEC initialization and just
	; do our own.  We use no EXEC routines.

	MVII	#$100,	R4	
	MVII	#$160,	R1	; $100...$25F. 
	CALL	FILLZERO	
	MVII	#$2C0,	R4	
	MVII	#$0A0,	R1	; $2C0...$35F. 
	CALL	FILLZERO	

	MVII	#CS0,	R4
	MVO@	R1,	R4	; Set color stack 0 to black
	MVO@	R1,	R4	; Set color stack 1 to black
	MVO@	R1,	R4	; Set color stack 2 to black
	MVO@	R1,	R4	; Set color stack 3 to black
	MVO@	R1,	R4	; Set border to black

	MOVR	PC,	R0	; (generate PC-relative address)
	ADDI	#INIT-$, R0	; Our initialization routine
	MVII	#ISRVEC, R4	; ISR vector
	MVO@	R0,	R4	; Write low half
	SWAP	R0		; 
	MVO@	R0,	R4	; Write high half


	EIS
@@spin:	DECR	PC

; Initializes the ISR, etc.  Gets everything ready to run.
; This is called via the ISR dispatcher, so it's safe to bang GRAM from
; here, too.
;
;  -- Zero out memory to get started
;  -- Set up variables that need to be set up here and there
;  -- Set up GRAM image
;  -- Drop into the main game state-machine.
;

INIT:   PROC
        DIS
        MVII    #$2F0,  R6      	; Reset the stack pointer

	
	MOVR	PC,	R0
        ADDI    #MAINISR-$,R0		; Point ISR vector to our ISR
        MVO     R0,     ISRVEC		; store low half of ISR vector
        SWAP    R0			; 
        MVO     R0,     ISRVEC+1	; store high half of ISR vector

        MVII    #SNDSTRM, R4            ; Point R4 to sound streams
        MVII    #2,     R0              ; Start off with hold count = 2
	MOVR	PC,	R1
	ADDI	#SONG-$,R1
	MVII	#$3FFF, R2
        MVO@    R0,     R4              ; Store hold count == 2
	MVO@	R1,	R4		; Store record ptr
	MVO@	R2,	R4		; Store stream mask
	MVO@	R2,	R4		; Store stream mask

        MVII    #2,     R0              ; Start off with hold count = 2
	SDBD
	MVII	#SONG2, R1
	MVII	#$3FFF, R2
        MVO@    R0,     R4              ; Store hold count == 2
	MVO@	R1,	R4		; Store record ptr
	MVO@	R2,	R4		; Store stream mask
	MVO@	R2,	R4		; Store stream mask

	; Ok, everything's ready to roll now.
@@spin:
        EIS

	MVI	SNDSTRM+1,  R0
	CMP	LAST1,	R0
	BEQ	@@not1
	MVO	R0,	LAST1
	CLRR	R1
	MVII	#20 + $200, R4
	CALL	DISPSTRM
@@not1:

	MVI	SNDSTRM+5,  R0
	CMP	LAST2,	R0
	BEQ	@@not2
	MVO	R0,	LAST2
	MVII	#4,	R1
	MVII	#80 + $200, R4
	CALL	DISPSTRM
@@not2:

	B	@@spin
        ENDP

;
; MAINISR
; This is the main interrupt service routine.  It is the sound player.
;
; PSG memory map
;
;       $1F4:$1F0       Channel A period (12 bits)
;       $1F5:$1F1       Channel B period (12 bits)
;       $1F6:$1F2       Channel C period (12 bits)
;       $1F7:$1F3       Envelope period (16 bits)
;
;       $1F8?           Channel enables
;       $1F9?           Noise period (5 bits)
;       $1FA?           Envelope mode
;
;       $1FB            Channel A volume
;       $1FC            Channel B volume
;       $1FD            Channel C volume
;
;       $1FE, $1FF      Hand controllers (not used for sound.)
;
; Sound is microprogrammed with a series of records of the following
; format:
;
;   HOLD COUNT, 1 Decle
;   +----+----+----+----+----+----+----+----+----+----+
;   | DURATION TO HOLD THIS FRAME |QUIT|ZERO|PCTL|BREF|
;   +----+----+----+----+----+----+----+----+----+----+
;
;   BACK REFERENCE, 1 Decle if BREF == 0
;   +----+----+----+----+----+----+----+----+----+----+
;   |        OFFSET IN DECLES RELATIVE TO $ - 2       |
;   +----+----+----+----+----+----+----+----+----+----+
;
;   CONTROL DECLE #1
;   +----+----+----+----+----+----+----+----+----+----+
;   | xx | xx |SND7|SND6|SND5|SND4|SND3|SND2|SND1|SND0|
;   +----+----+----+----+----+----+----+----+----+----+
;
;   CONTROL DECLE #2
;   +----+----+----+----+----+----+----+----+----+----+
;   | xx | xx | 0  |LINK|SNDD|SNDC|SNDB|SNDA|SND9|SND8|
;   +----+----+----+----+----+----+----+----+----+----+
;
;
;   DATA BYTES, 1 Decle for each SNDx bit set
;   +----+----+----+----+----+----+----+----+----+----+
;   | xx | xx |          DATA                         |
;   +----+----+----+----+----+----+----+----+----+----+
;
;   LINK ADDRESS (if LINK set), 2 Decles, SDBD format
;   +----+----+----+----+----+----+----+----+----+----+   
;   | xx | xx |     LINK ADDRESS LOW                  |  
;   +----+----+----+----+----+----+----+----+----+----+      
;   +----+----+----+----+----+----+----+----+----+----+                      
;   | xx | xx |     LINK ADDRESS HIGH                 |  
;   +----+----+----+----+----+----+----+----+----+----+ 
;   
;
; The bits SND0 ... SNDD report whether bytes are coded for each of the
; fourteen sound registers on the PSG.
;
; The QUIT, ZERO, PCTL, BREF and LINK flags control the following behavior:
;
;   QUIT   When set, this terminates the sound given sound microprogram
;   ZERO   When set, this tells the sound processor that no bytes are
;          coded for this frame since all were zeros.  This causes the
;          processor to write a pattern of zeros guided by the control word.
;   PCTL   When set, this tells the sound processor to use the previous
;          control word.  The previous control word is not defined when
;          the stream is started, but it is defined across a link.
;   BREF   When set, this frame is coded only as a back-reference.
;          All control decles and data bytes come from the back reference.
;          Note: When using a back-reference, the QUIT, ZERO, PCTL, and
;          DUR settings are used from the _current_ frame!
;   LINK   When set, this tells the sound processor that this frame
;          is followed by the address for the next frame.  Otherwise,
;          the sound processor assumes that the next frame is after this one.
;   
;
; Up to two sound programs may be running at once.  It is up to the
; programmer to ensure that they do not try to control the same channels
; at once.  The intended use is to have music playing on one stream
; and sound effects on a second stream.
;
; The following state is kept for each active sound stream in 16-bit
; memory:
;
;     HOLDCNT    -- Number of ticks left for current record times 2
;                   If the LSB of HOLDCNT is set, stream processing stops
;                   after this record.
;     PREVCTL    -- Previous control word.
;     NEXTREC    -- Pointer to the next stream record.
;     STRMMSK    -- Stream channel mask.  This is used to prevent a
;                   stream from updating certain channel parameters and
;                   is most useful when, say, a sound-effect needs to
;                   borrow a channel briefly.
;
;
; NOTE: Bit 15 of control word (eg. bit 7 of second word) MUST BE 0.
;       Otherwise, copy loop will loop forever.  Also, bit 14 of stream
;       channel mask MUST BE 0, otherwise the copy loop will write part
;       of the link address to one of the hand controller addresses.  :-)
;



MAINISR         PROC

        PSHR    R5                      ; Save return address
	DIS

        ;; Hit the VBLANK handshake now
        MVO     R0,     $20             ; Allow STIC to display this frame.

	; Update music
        MVII    #SNDSTRM, R4            ; Point to the correct stream struct
	MVII	#PSG1, R0
	MVO	R0,	PSG
	CALL	SOUNDUPDATE
	MVII	#PSG2, R0
	MVO	R0,	PSG
        MVII    #SNDSTRM+4, R4          ; Point to the correct stream struct
	CALL	SOUNDUPDATE
	EIS
	PULR	PC
	
SOUNDUPDATE:
	PSHR	R5

        MVI@    R4,     R1              ; Load tick/hold count
        SUBI    #2,     R1              ; Count down the hold count
        BMI     @@inactive              ; If count goes negative, channel is
                                        ;    inactive.
        BEQ     @@nextrec               ; If count goes to zero, load next
                                        ;    sound record into the channel

        DECR    R4                      ; Otherwise, store the updated count.
        MVO@    R1,     R4

@@inactive:
        PULR    PC                      ; Return.

@@backup:
        DECR    R4
@@nextrec:
        MVI@    R4,     R5              ; Load the stream record pointer
        MVI@    R5,     R2              ; Load hold tick-count from stream
	SARC	R2,	1		; See if this is a back-reference
	BNC	@@notbref
	DECR	R4
	MVI@	R5,	R3		; Get reverse offset
	MVO@	R5,	R4		; Store updated stream rec pointer
	ADDI	#3,	R3		; Make BREF relative to just before rec
	SUBR	R3,	R5		; Subtract relative offset
@@notbref:
        SARC    R2,     1               ; See if we have new control word.
        BC      @@prevctl
        SDBD
        MVI@    R5,     R1              ; Load new control-word from stream
        MVO@    R1,     R4              ; Store updated ctrl-word to str info
        INCR    R7                      ; Skip the following MVI@.
@@prevctl:
        MVI@    R4,     R1              ; Load previous ctrl-word from str info

        MVI@    R4,     R0              ; Load the stream channel mask
        SUBI    #4,     R4              ; Point back at hold tick-count
        SARC    R2,     1               ; See if we are writing a zero-record
        BNEQ    @@notzero
        ; If this was a zero hold count, set up so that we will process
        ; next record.  This is almost like tail-recursion.
        MVII    #@@backup, R3
        PSHR    R3
@@notzero:
        MVO@    R2,     R4              ; Store new tick count to stream


        ; We generate a write mask by ANDing the stream channel mask
        ; with the control word.  The result is a bit-mask describing
        ; which PSG registers we'll actually write to.
        ANDR    R1,     R0              ; Generate channel write mask
        MVI     PSG,    R2              ; Point to the PSG for writing.

        BC      @@zerorec               ; If we're writing a zero record,
                                        ; don't read any bytes from stream.

        COMR    R0
        SARC    R1,     1       ; 6     ; See if we have a coded word.
@@loop:
        BNC     @@noread        ; 9 7   ; If not, skip the read.
@@read:
        MVI@    R5,     R3      ; 8     ; ... otherwise read word from stream
        SARC    R0,     1       ; 6     ; See if we are to write this word.
        ADCR    R7              ; 6
        MVO@    R3,     R2      ; 8     ; ... otherwise write to the PSG
        INCR    R2              ; 6     ; Always update the PSG pointer.
                                 
        SARC    R1,     1       ; 6     ; See if we have a coded word.
        BNEQ    @@loop          ; 9 7   ; Yes: Keep looping!
        BC      @@read          ;       ; affects last bit ONLY.

@@finish:

	MVI	PSG,	R0
	ADDI	#15,	R0
        CMPR    R0,     R2              ; Did we have a link pointer in this
                                        ; record?
        BLT     @@nolink                ; If not, then don't grab one here.
        SWAP    R3
        XOR@    R5,     R3              ; Grab link pointer from stream
        SWAP    R3                      ; Put bytes in right order
        MOVR    R3,     R5              ; Move link pointer to record pointer
	B	@@waslink		; Skip the back-ref test.
@@nolink:
	CMP@	R4,	R5		; Detect if this was a back-ref
	BLE	@@return		; If it was, our pointer didn't advance
	DECR	R4			; Else, prepare to store new rec ptr
@@waslink:
        MVO@    R5,     R4              ; Store record ptr to stream info
@@return:
        PULR    PC                      ; Return

@@noread:
        INCR    R2              ; 6
        SARC    R0,     1       ; 6
        SARC    R1,     1       ; 6
        BNEQ    @@loop          ; 9 7
        BC      @@read          ;      ; affects last bit ONLY.
        B       @@finish
        
                                ; 56  coded and written
                                ; 48  coded, not written
                                ; 36  not coded, not written
@@zerorec:
        CLRR    R3
        SARC    R0,     1       ; 6     ; See if we are to write this word.
@@zloop:
        BNC     @@nozero        ; 9 7   ; Skip write if not in write mask
        MVO@    R3,     R2      ; 8     ; ... otherwise write to the PSG
@@nozero:
        INCR    R2              ; 6     ; Always update the PSG pointer.
        SARC    R0,     1       ; 6     ; See if we are to write this word.
        BNEQ    @@zloop         ; 9 7   ; Yes: Keep looping!
        BNC     @@zlink
        MVO@    R3,     R2

        ; We need to handle link records specially here.
@@zlink:
        SLLC    R1,     2               ; Check bit 14.  
        BNOV    @@nolink                ; If set, we have link, else, we don't
        SDBD
        MVI@    R5,     R3              ; Load the link address
        MVO@    R3,     R4              ; Store record ptr to stream info

        PULR    PC
        
        ENDP
;
; FILLZERO
; Fills memory with zeros
;
; FILLMEM
; Fills memory with a constant
;
; INPUTS:
; R0 -- Fill value (FILLMEM only)
; R1 -- Number of words to fill
; R4 -- Start of fill area
; R5 -- Return address
;
; OUTPUTS:
; R0 -- Zeroed (if FILLZERO)
; R1 -- Zeroed
; R4 -- Points to word after fill area
;
FILLZERO        PROC
        CLRR    R0
FILLMEM
        MVO@    R0,     R4
        DECR    R1
        BNEQ    FILLMEM
        JR      R5
        ENDP

; DISPSTRM
; Displays stream information to screen:
;
;     01234567890123456789
;     PPPP           WWWW
;     AAA A  BBB B  CCC C
;
; INPUTS:
; R1 -- Stream number * 4 (eg. 0 or 4)
; R4 -- Display Pointer
; R5 -- Return address.
;
; OUTPUTS:
; R0..R5 trashed.
; Uses $310..$312 for temp storage

DISPSTRM	PROC
@@tmp0		EQU	$310
@@tmp1		EQU	$311
@@tmp2		EQU	$312

	PSHR	R5

	MVO	R1,	@@tmp0		; Save our stream number
	ADDI	#SNDSTRM+1, R1		; Point to stream record pointer
	
	MVO	R1,	@@tmp1		; Save pointer
	MVI@	R1,	R0		; Get stream record pointer

	MVII	#6,	R3		; Display in yellow
	CALL	HEX16			; Put it onscreen.

	MVI	@@tmp1,	R1		; Restore our pointer
	INCR	R1			; Point to stream control word	
	MVI@	R1,	R0		; Get stream control word

	MVII	#1,	R3		; Display in blue
	ADDI	#11,	R4		; Move to right side of screen
	CALL	HEX16			; Put it onscreen.
	INCR	R4			; Move to next line

	MVI	@@tmp0,	R3		; Get our stream number
	MOVR	R3,	R1		; Copy it to R1
	SLR	R3,	1		; R3 = 0 or 2
	ADDR	R3,	R1		; R1 = 0 or 6
	ADDI	#PSV1,	R1		; R1 points to PSG save area 1 or 2
	SLR	R3,	1		; R3 = 0 or 1
	SWAP	R3,	1		; R3 = $000 or $100
	XORI	#$1F0,	R3		; R3 = $1F0 or $0F0 --> PSG 1 or 2

	INCR	PC			; Fall through into DISPPSG below
					; by skipping the PSHR R5
	ENDP

; DISPPSG
; Displays a PSG's state on the screen:
;
;     AAA A  BBB B  CCC C
;
; INPUTS:
; R1 -- Prev PSG state area pointer
; R3 -- PSG pointer
; R4 -- Screen pointer
; R5 -- Return address.
;
; OUTPUTS:
; R0..R3, R5 -- Trashed
; R4 -- Points two chars past last char written
; Uses $110..$112 for temp storage


DISPPSG:	PROC
@@tmp0		EQU	$310
@@tmp1		EQU	$311
@@tmp2		EQU	$312

	PSHR	R5

	MVII	#3,	R2	; Three channels, A, B, and C.

	ADDI	#4,	R3	; Point to channel period MSBs

@@chanlp:
	MVI@	R3,	R0	; Get MSBs
	SWAP	R0,	1
	SUBI	#4,	R3	
	XOR@	R3,	R0 	; Merge with LSBs
	ADDI	#11,	R3	; Point to this channel's volume reg
	
	SLL 	R0,	2	; Shift value left 4
	SLL 	R0,	2

	MVO	R3,	@@tmp0	; Save our PSG pointer
	MVO	R2,	@@tmp1 	; Save our loop counter

	MVII	#7,	R3	; Default to white
	CMP@	R1,	R0	; Did this value change?
	BEQ	@@nope1
	MVII	#6,	R3	; If yes, display in yellow
@@nope1:

	MVO@	R0,	R1	; Save updated period to PSG state area
	INCR	R1		; Point to next entry
	MVO	R1,	@@tmp2 	; Save our PSG state area pointer

	CALL	HEX12		; Display the period
	INCR	R4

	MVI	@@tmp0,	R3	; Restore PSG pointer
	MVI	@@tmp2, R1	; Restore PSG state area pointer

	MVI@	R3,	R0	; Get channel's volume
	ANDI	#$F,	R0	; Limit to 0..15
	SUBI	#6,	R3	; Point to next channel's MSBs
	MVO	R3,	@@tmp0	; Save our PSG pointer

	SWAP	R0,	1	; Put volume in four MSBs of word
	SLL	R0,	2
	SLL	R0,	2

	MVII	#7,	R3	; Default to white
	CMP@	R1,	R0	; Did this value change?
	BEQ	@@nope2
	MVII	#6,	R3	; If yes, display in yellow
@@nope2:
	MVO@	R0,	R1	; Save updated volume to PSG state area
	INCR	R1		; Point to next entry
	MVO	R1,	@@tmp2	; Save our PSG state area pointer

	CALL	HEX4 		; Display the volume
	ADDI	#2, 	R4	; Move to next channel's field.

	MVI	@@tmp0,	R3	; Restore PSG pointer
	MVI	@@tmp2, R1	; Restore PSG state area pointer
	MVI	@@tmp1, R2	; Restore loop counter
	DECR	R2		; Do next channel if there is one.
	BNEQ	@@chanlp

	PULR	PC
	ENDP

	
;
; HEX16 -- Display a 4-digit hex number on the screen
; HEX12 -- Display a 3-digit hex number on the screen (MSBs)
; HEX8  -- Display a 2-digit hex number on the screen (MSBs)
; HEX4  -- Display a 1-digit hex number on the screen (MSBs)
;
; INPUTS:
; R0 -- Hex number
; R4 -- Screen offset
; R3 -- Color mask
;
; OUTPUTS:
; R0 -- Rotated by (4 * number of digits + 3)
; R1 -- trashed
; R2 -- zeroed
; R4 -- points just to right of string
HEX16           PROC
	MVII	#4,	R2
	B	@@start
HEX12:
	MVII	#3,	R2
	B	@@start
HEX8:
	MVII	#2,	R2
	B	@@start
HEX4:
	MVII	#1,	R2

@@start:
        ; Rotate R0 left by 3
        MOVR    R0,     R1
        SLLC    R1,     2
        RLC     R0,     2
        SLLC    R1,     1
        RLC     R0,     1
@@loop:
        ; Rotate R0 left by 4
        MOVR    R0,     R1
        SLLC    R1,     2
        RLC     R0,     2
        SLLC    R1,     2
        RLC     R0,     2
        MOVR    R0,     R1
        ANDI    #$78,   R1
        CMPI    #$50,   R1
        BLT     @@digit
        ADDI    #$38,   R1
@@digit:
        ADDI    #$80,   R1
        XORR    R3,     R1
        MVO@    R1,     R4

        DECR    R2
        BNE     @@loop

        JR      R5
        ENDP

;------------------------------------------------------------------------------
OFFSET:
;------------------------------------------------------------------------------

SONG:	PROC
	include "song.asm"
@@loop:
	DECLE	$00C
	WORD	$3FFF

	ENDP

SONG2:	PROC
	include "song2.asm"
@@loop:
	DECLE	$00C
	WORD	$3FFF

	ENDP

